import os
import time

import pytest
import requests
from web3 import Web3

from src.pybundlr import pybundlr

def test_balance_arweave():
    ar_address = "Ry2bDGfBIvYtvDPYnf0eg_ijH4A1EDKaaEEecyjbUQ4"
    bal = pybundlr.balance(ar_address, "arweave") #not AR
    assert bal > 0

    
def test_balance_ethereum():
    eth_address = "0x7BA3d8551A6f2C70a5d47bb448BcF7EF69661822"
    bal = pybundlr.balance(eth_address, "ethereum") #not ETH
    assert bal >= 0


def test_balance_matic():
    eth_address = "0x7BA3d8551A6f2C70a5d47bb448BcF7EF69661822"
    
    #it uses "matic"
    bal = pybundlr.balance(eth_address, "matic")
    assert bal >= 0

    #it doesn't recognize "polygon"
    with pytest.raises(ValueError) as e_info:
        pybundlr.balance(eth_address, "polygon")
    assert "Unknown/Unsupported currency polygon" in str(e_info)


def test_balance_foo():
    address = "Ry2bDGfBIvYtvDPYnf0eg_ijH4A1EDKaaEEecyjbUQ4"
    with pytest.raises(ValueError) as e_info:
        pybundlr.balance(address, "foo")
    assert "Unknown/Unsupported currency foo" in str(e_info)


def test_fund_and_withdraw():
    eth_private_key = os.getenv('REMOTE_TEST_PRIVATE_KEY1')
    eth_address = pybundlr.eth_address(eth_private_key)

    amt_fund_wei = 3
    pybundlr.fund(amt_fund_wei, "matic", eth_private_key)

    amt_withdraw_wei = 2

    #give time for deposit to go through, if needed
    for i in range(10):
        bal_wei = pybundlr.balance(eth_address, "matic")
        if bal_wei >= amt_withdraw_wei:
            break
        print(f"Not enough funds yet, so sleep. bal_wei={bal_wei}")
        time.sleep(1.0)        

    try:
        pybundlr.withdraw(amt_withdraw_wei, "ethereum", eth_private_key)
    except ValueError as e_info:
        #let it pass if insufficient balance, otherwise error gets raised
        #why: the funding above doesn't seem to make it through (FIXME)
        if not "Insufficient Balance" in str(e_info):
            raise ValueError(e_info)


def test_fund_fail_no_eth():
    #randomly create a new account, which therefore has 0 eth
    account = Web3().eth.account.create()
    eth_private_key = account.key.hex()[2:]

    #watch funding fail
    amt_wei = 3
    with pytest.raises(ValueError) as e_info:
        pybundlr.fund(amt_wei, "matic", eth_private_key)
        
    assert "Can't fund" in str(e_info) or "insufficient funds" in str(e_info)


def test_price(tmp_path):
    num_bytes = 10
    amt_wei = pybundlr.price(num_bytes, "matic")
    assert amt_wei > 100000000  # example observed number: 24275774772


def test_upload(tmp_path):
    content_in = "mycontent"
    f = tmp_path / "myfile.txt"
    f.write_text(content_in)
    file_name = str(f)

    num_bytes = os.stat(file_name).st_size

    eth_private_key = os.getenv('REMOTE_TEST_PRIVATE_KEY1')
    p = pybundlr.price(num_bytes, "matic")
    
    amt_wei = p * 2 # the 2x is for safety margin, since price fluctuates
    pybundlr.fund(amt_wei, "matic", eth_private_key)
    
    url = pybundlr.upload(file_name, "matic", eth_private_key)
    assert "https://arweave.net/" in url, url

    result = requests.get(url)
    content_out = result.text

    assert content_in == content_out, content_out


def test_fund_and_upload(tmp_path):
    content_in = "mycontent"
    f = tmp_path / "myfile.txt"
    f.write_text(content_in)
    file_name = str(f)

    eth_private_key = os.getenv('REMOTE_TEST_PRIVATE_KEY1')
    
    pybundlr.fund_and_upload(file_name, "matic", eth_private_key)
    
    url = pybundlr.upload(file_name, "matic", eth_private_key)
    assert "https://arweave.net/" in url, url

    result = requests.get(url)
    content_out = result.text

    assert content_in == content_out, content_out


def test_eth_address():
    eth_private_key = os.getenv('REMOTE_TEST_PRIVATE_KEY1')
    eth_address = pybundlr.eth_address(eth_private_key)
    assert eth_address[:2] == "0x"
    assert len(eth_address) == 42


def test_bal_on_ethereum():
    #address with known nonzero eth
    eth_address = "0x7BA3d8551A6f2C70a5d47bb448BcF7EF69661822"
    assert pybundlr.bal_on_ethereum(eth_address) > 0


def test_bal_on_matic():
    #address with known nonzero eth
    eth_address = "0x7BA3d8551A6f2C70a5d47bb448BcF7EF69661822"
    assert pybundlr.bal_on_matic(eth_address) > 0
    



import os
import re
import subprocess

from enforce_typing import enforce_types
import web3
from web3 import Web3

BUNDLR_NODE_URL = "https://node1.bundlr.network"

#picked arbitrarily from https://chainlist.org/chain/1
ETH_NODE_URL = "https://eth-mainnet.nodereal.io/v1/1659dfb40aa24bbb8153a677b98064d7"
MATIC_NODE_URL = "https://polygon-rpc.com/"


#==========================================================================
#functions that are 1:1 with bundlr CLI

@enforce_types
def balance(address:str, currency:str) -> int:
    """
    Gets the specified user's balance for the current Bundlr node
    
    Parameters:
      address - balance of address for the given currency at bundlr node
      currency - Eg "arweave" (AR), "ethereum" (ETH), "matic" (MATIC)

    Returns:
      balance - amount held of the target currency, in base unit (winston, wei)
    """
    cmd = f"bundlr balance {address} -c {currency} -h {BUNDLR_NODE_URL}"
    stdout = _run_cmd(cmd)
    bal_s = re.findall("\d+", stdout)[0]
    bal = int(bal_s)
    return bal


@enforce_types
def fund(amount:int, currency:str, private_key: str) -> str:
    """
    Funds your account with the specified amount of atomic units

    Parameters:
      amount - how much to fund. In base units (winston, wei, ..)
      currency - Eg "arweave" (AR), "ethereum" (ETH), "matic" (MATIC)
      private_key - private key, or path to json with arweave wallet
    """
    if currency == "ethereum":
        addr = eth_address(private_key)
        b = bal_on_ethereum(addr)
        if b < amount:
            raise ValueError(f"Can't fund {amount} wei: balance is {b}")

    cmd = f"bundlr fund {amount} -c {currency} -h {BUNDLR_NODE_URL} " \
        f"-w {private_key} --no-confirmation"
    stdout = _run_cmd(cmd)


@enforce_types
def withdraw(amount:int, currency:str, private_key: str) -> str:
    """
    Sends a fund withdrawal request

    Parameters:
      amount - how much to withdraw. In base units (winston, wei, ..)
      currency - Eg "arweave" (AR), "ethereum" (ETH), "matic" (MATIC)
      private_key - private key, or path to json with arweave wallet
    """
    cmd = f"bundlr withdraw {amount} -c {currency} -h {BUNDLR_NODE_URL} " \
        f"-w {private_key} --no-confirmation"
    stdout = _run_cmd(cmd)


@enforce_types
def price(num_bytes: int, currency:str) -> int:
    """
    Check how much of a specific currency is required for an upload of num_bytes

    Parameters:
      num_bytes -- how many bytes. E.g. via os.stat(file_name).st_size
      currency -- Eg "arweave" (AR), "ethereum" (ETH), "matic" (MATIC)

    Returns:
      amt - price to upload, in base unit (winston, wei)
    """
    cmd = f"bundlr price {num_bytes} -c {currency} -h {BUNDLR_NODE_URL} "
    stdout = _run_cmd(cmd)

    #e.g. stdout = "Price for 10 bytes in ethereum is 24294303017 wei ..."
    amt_s = re.findall("\d+", stdout)[1]
    amt = int(amt_s)
    return amt


@enforce_types
def upload(file_name:str, currency:str, private_key:str) -> str:
    """
    Uploads a specified file

    Parameters:
      file_name -- path to file
      currency -- Eg "arweave" (AR), "ethereum" (ETH), "matic" (MATIC)
      private_key - private key, or path to json with arweave wallet

    Returns:
      url - location on arweave network where file is now stored
    """
    cmd = f"bundlr upload {file_name} -c {currency} -h {BUNDLR_NODE_URL} " \
        f"-w {private_key} --no-confirmation"
    stdout = _run_cmd(cmd)

    #e.g. "Uploaded to https://arweave.net/PJVOHDHjYrTXQJQg9UlgfKxgV2dUspc"
    url = stdout.split()[-1] 
    return url

#==========================================================================
#functions that combine >1 CLI calls

def fund_and_upload(file_name:str, currency:str, private_key:str) -> str:
    """
    Uploads a specified file. Funds first if needed.
    For "ethereum" or "matic" only, for now.

    Parameters:
      file_name -- path to file
      currency -- "ethereum" (ETH), "matic" (MATIC)
      private_key - private key, or path to json with arweave wallet

    Returns:
      url - location on arweave network where file is now stored
    """
    #we don't yet have a convenient conversion from AR private key -> addr
    assert currency in ["ethereum", "matic"], "only currently support evm"
    
    num_bytes = os.stat(file_name).st_size

    p = price(num_bytes, currency)
    
    amt_wei = p * 2 # the 2x is for safety margin, since price fluctuates

    addr = eth_address(private_key)
    bal_wei = balance(addr, currency)
    if bal_wei < amt_wei:
        fund(amt_wei, currency, private_key)

    url = upload(file_name, currency, private_key)

    return url


#==========================================================================
#helper method

def _run_cmd(cmd:str):
    print(f"\nRUN COMMAND: {cmd}")
    args = cmd.split()
    completed_process = subprocess.run(args, capture_output=True, check=True)

    stdout = completed_process.stdout.decode("ascii")
    stderr = completed_process.stderr.decode("ascii")

    if "error" in stderr.lower() or stdout == "":
        print(stderr)
        raise ValueError(stderr)
    
    print(stdout)
    return stdout


#==========================================================================
#eth convenience functions

def eth_address(eth_private_key:str) -> str:
    account = web3.eth.Account.from_key(eth_private_key)
    return account.address


def bal_on_ethereum(eth_address:str) -> int:
    """
    Returns ether balance on Ethereum mainnet
    
    Parameters:
      eth_address - address on eth mainnet

    Returns:
      bal - ether balance, denominated in wei
    """
    return _w3_ethereum().eth.get_balance(eth_address)


def bal_on_matic(eth_address:str) -> int:
    """
    Returns ether balance on Matic
    
    Parameters:
      eth_address - address on eth mainnet

    Returns:
      bal - ether balance, denominated in wei
    """
    return _w3_matic().eth.get_balance(eth_address)

def _w3_ethereum():
    """Returns - web3 object pointed to Ethereum mainnet"""
    return Web3(Web3.HTTPProvider(ETH_NODE_URL))

def _w3_matic():
    """Returns - web3 object pointed to Matic (Polygon) mainnet """
    return Web3(Web3.HTTPProvider(MATIC_NODE_URL)) 


